#!/usr/bin/env python3
from coyote import (
    Coyote,
    ensure_enddef,
    get_field_metadata,
    group_comm_rank,
    group_comm_size,
    init,
    run,
    start_datetime,
)

from ._data_handler import DataHandler
from ._distribute_work import distribute_work
from ._grids import def_grid, grid_id
from ._zarr_utils import get_var_group, get_var_parent_group, get_time_axis
from .loco import LocoServer

from argparse import ArgumentParser
from itertools import chain, groupby

import json
import logging
import numpy as np
import zarr


def main():
    parser = ArgumentParser()
    # zarr.copy_store (used with loco) does not work with zarr.open_consolidated
    parser.add_argument(
        "datasets",
        type=str,
        nargs="+",
        help="URL of the dataset. To be passed to zarr.open.",
    )
    parser.add_argument(
        "--log-level",
        default=logging.INFO,
        type=lambda x: getattr(logging, x),
        help="Configure the logging level.",
    )
    parser.add_argument(
        "--process-group",
        default="hiopy",
        help="A string determining the process group for distributing the work",
    )
    parser.add_argument("--loco", action="store_true", help="Run in LOCO mode")
    parser.add_argument(
        "--loco-host", type=str, default="0.0.0.0", help="Host address of the LOCO webserver"
    )
    parser.add_argument("--loco-port", type=int, default=8080, help="Port of the LOCO webserver")
    parser.add_argument(
        "--zarr-version",
        type=int,
        default=None,
        help="Version of the zarr store. "
        "(This option becomes obsolete if zarr.open can infer that information from the store)",
    )
    parser.add_argument(
        "--nthreads", type=int, default=1, help="Number of threads to use to proceed data"
    )
    args = parser.parse_args()
    args.datasets = [zarr.open(ds, zarr_version=args.zarr_version) for ds in args.datasets]

    if args.loco:
        assert len(args.datasets) == 1, "Loco only supports reading from 1 dataset"
        loco_store = zarr.MemoryStore()
        zarr.copy_store(args.datasets[0].store, loco_store)
        zarr.consolidate_metadata(loco_store)
        loco_server = LocoServer(loco_store, args.loco_host, args.loco_port)
        args.datasets = [zarr.open(store=loco_store)]

    init(group_name=args.process_group, nthreads=args.nthreads)
    rank, size = group_comm_rank(), group_comm_size()

    logging.basicConfig(
        level=args.log_level, format=f"{args.process_group}({rank}/{size}): %(message)s"
    )

    # find all variables considered to be written in the input datasets:
    def collect_data_vars(group, root):
        for _name, item in group.arrays():
            if "hiopy::enable" in item.attrs and item.attrs["hiopy::enable"]:
                item.root = root
                yield item
        for _name, item in group.groups():
            yield from collect_data_vars(item, root)

    all_data_vars = list(chain(*[collect_data_vars(z, z) for z in args.datasets]))
    logging.info(f"Found {len(all_data_vars)} variables")
    if len(all_data_vars) == 0:
        raise RuntimeError("No variables found by the hiopy worker.")

    # group the variables by the crs grid_mapping.
    # This is used to distribute them through the processes and create the coyote instances
    grouped_data_vars = {
        gid: list(variables)
        for gid, variables in groupby(sorted(all_data_vars, key=grid_id), key=grid_id)
    }
    distributed_data_vars = distribute_work(grouped_data_vars, group_comm_size())

    if args.loco:
        loco_server.write_nginx_config(distributed_data_vars, group_comm_rank())

    my_data_vars = distributed_data_vars[group_comm_rank()]
    # my_data_vars list of tuples: (gid, data_vars, slice)

    if group_comm_rank() == 0:
        logging.debug(distributed_data_vars)

    frac_masks = {}  # the numpy array must keep alive until the end of the program

    coyote_instances = {
        gid: Coyote(f"{args.process_group}_{gid}") for gid, data_vars, chunk_slice in my_data_vars
    }

    data_handlers = []

    for gid, data_vars, chunk_slice in my_data_vars:
        coyote = coyote_instances[gid]
        # all vars in data_vars define the same grid
        def_grid(coyote, data_vars[0], chunk_slice)

        for v in data_vars:
            # compute timestep
            var_group = get_var_group(v)
            time_coordinate = get_time_axis(v)

            dt = time_coordinate[1] - time_coordinate[0]
            # compute time start index
            t0 = (
                np.datetime64(start_datetime())
                - np.datetime64(var_group["time"].attrs["units"][len("seconds since ") :])
            ) / np.timedelta64(1, "s") + dt
            t0_idx = np.searchsorted(var_group["time"], t0)
            assert var_group["time"][t0_idx] == t0, f"start_datetime {t0} not found in time axis \
                                    at index {t0_idx} which has value {var_group['time'][t0_idx]}"

            # see YAC_REDUCTION_TIME_NONE etc. (TODO: pass constants through coyote)
            time_methods2yac = {"point": 0, "sum": 1, "mean": 2, "min": 3, "max": 4}

            collection_size = v.shape[1] if len(v.shape) == 3 else 1
            src_name = v.attrs.get("hiopy::src_name", default=v.basename)
            if "hiopy::yac_source" in v.attrs:
                src_comp, src_grid = v.attrs["hiopy::yac_source"]
            else:
                assert "hiopy::parent" in var_group.attrs, f"No source for field {v.name} specified"
                parent_group = get_var_parent_group(v)
                source_var = parent_group[v.basename]
                src_name = source_var.name
                source_var_gid = grid_id(source_var)
                src_comp = src_grid = f"{args.process_group}_{source_var_gid}"
            time_method = v.attrs.get("hiopy::time_method", "point")
            nnn = v.attrs.get("hiopy::nnn", 1)
            frac_mask_name = v.attrs.get("hiopy::frac_mask", None)
            frac_mask = []
            if frac_mask_name is not None:
                if frac_mask_name not in frac_masks:
                    frac_masks[frac_mask_name] = np.array(var_group[frac_mask_name][chunk_slice])
                frac_mask = frac_masks[frac_mask_name]

            logging.info(
                f"registering {v.name} with "
                f"{collection_size=}, "
                f"{src_comp=}, "
                f"{src_grid=}, "
                f"{src_name=}, "
                f"{time_method=}"
            )
            data_handlers.append(DataHandler(v, t0_idx, chunk_slice))

            if args.loco:
                data_handlers[-1].set_loco_server(loco_server)

            coyote.add_field(
                src_comp,
                src_grid,
                src_name,
                data_handlers[-1],
                f"PT{dt}S",
                collection_size,
                yac_time_reduction=time_methods2yac[time_method],
                interpolation_stack_yaml=f"- nnn:\n    n: {nnn}",
                frac_mask=frac_mask,
                pass_through_name=v.name,
            )

    def get_source_triple(v):
        var_group = get_var_group(v)
        if "hiopy::parent" in var_group.attrs:
            pgroup = get_var_parent_group(v)
            return get_source_triple(pgroup[v.basename])
        elif "hiopy::yac_source" in v.attrs:
            src_comp, src_grid = v.attrs["hiopy::yac_source"]
            src_field = v.attrs.get("hiopy::src_name", v.basename)
            return src_comp, src_grid, src_field
        else:
            raise RuntimeError("Invalid attributes: " + str(dict(v.attrs)))

    ensure_enddef()
    if group_comm_rank() == 0:
        for v in all_data_vars:
            if "hiopy::copy_metadata" in v.attrs:
                comp, grid, field = get_source_triple(v)
                md_str = get_field_metadata(comp, grid, field)
                metadata = json.loads(md_str)
                logging.debug(f"Found metadata for {field}: {metadata}")
                for key, value in metadata["cf"].items():
                    v.attrs[key] = value
                del v.attrs["hiopy::copy_metadata"]  # copy only once

        # re-consolidate the newly updated metadata
        for ds in args.datasets:
            zarr.consolidate_metadata(ds.store)

    run()

    for dh in data_handlers:
        dh.flush()


if __name__ == "__main__":
    main()
