#!/bin/bash
# More safety by turning some bugs into errors.
# Without `errexit` you don’t need ! and can replace
# ${PIPESTATUS[0]} with a simple $?, but I prefer safety.
set -o errexit -o pipefail -o noclobber

REPO_TAG_FYAML="v0.9"
REPO_TAG_CFITSIO="cfitsio4_5_0_20240826"
REPO_TAG_HEALPIX="trunk" # because its only a mirror. version control is done else where.
REPO_TAG_YAXT="release-0.11.3"
REPO_TAG_YAC="release-3.4.0_p2"
REPO_TAG_PYBIND11="v2.13"

function usage {
    echo "This script is intendet to build YACO with its dependencies on Levante/LUMI"
    echo "Each checked out branch/version of the repositories will get their own repo/install/build folder"
    echo ""
    echo "--- Workflow --------------------------------------------------------------"
    echo "   build-dependencies -j <thread_count>"
    echo "          starts the building process with given number of threads"
    echo ""
    echo "--- Options ---------------------------------------------------------------"
    echo "   -j                              Threadcount"
    echo "   --dependency <name>/all         Set the library name to be built eg: healpix"
    echo "                                   valid is one of [fyaml, cfitsio, healpix, yac, yaxt]"
    echo "                                   defaults to building all depedencies"
}


function clone_to {
    if [ -d "$2" ]
    then
        cd "$2"
        echo "repo already loaded: skipping git clone"
    else
        if [ -z "$3" ]; then
            echo "Using default branch"
            git clone --depth=1 --recursive "$1" "$2"
        else
            git clone --depth=1 --recursive --branch "$3" "$1" "$2"
        fi
    fi
}

# invoke with the specified FC, CC and CXX if they are present as variables
function invoke_build_script_generation {
    if [ -n "${CC}" ] || [ -n "${CXX}" ] || [ -n "${FC}" ]; then
    echo "env CC=${CC:-$CC} CXX=${CXX:-$CXX} FC=${FC:-$FC} "$@""
	env CC=${CC:-$CC} CXX=${CXX:-$CXX} FC=${FC:-$FC} "$@"
    else
        "$@"
    fi
}


function install_via_cmake {
    REPO_NAME=$1
    shift
    REPO_TAG=$1
    shift
    GIT_URL=$1
    shift

    local REPO_TAG=${REPO_TAG/\//-}
    local VERSIONED_FOLDER=$REPO_NAME-$REPO_TAG
    local REPO_PATH=$GIT_REPOS/$VERSIONED_FOLDER
    local BUILD_FOLDER=$REPO_PATH/build

    clone_to "$GIT_URL" "$REPO_PATH" "$REPO_TAG"
    mkdir -p "$BUILD_FOLDER"
    cd "$BUILD_FOLDER"

    invoke_build_script_generation cmake "$REPO_PATH" -DCMAKE_INSTALL_LIBDIR="${INSTALL_DIR}/lib" -DCMAKE_INSTALL_PREFIX="$INSTALL_DIR" -DCMAKE_POSITION_INDEPENDENT_CODE=ON -DCMAKE_PREFIX_PATH="$INSTALL_DIR" "$@"
    make -j "$THREADS"
    make install
}

function check_and_install {
  PKG_NAME=$1
  if ! pkg-config --exists $PKG_NAME; then
    echo "Package $PKG_NAME not found. Installing it now."
    install_${PKG_NAME}
  else
    echo "Package $PKG_NAME is already installed. Skipping this step."
  fi
}

function install_libfyaml {
    GIT_URL="https://github.com/pantoniou/libfyaml.git"
    REPO_NAME="libfyaml"

    install_via_cmake "$REPO_NAME" "$REPO_TAG_FYAML" "$GIT_URL"
}

function install_cfitsio {
    REPO_NAME="cfitsio"
    VERSIONED_FOLDER=$REPO_NAME-$REPO_TAG_CFITSIO
    REPO_PATH=$GIT_REPOS/$VERSIONED_FOLDER
    LIBCFITSIO_BUILD_FOLDER=$REPO_PATH/build
    clone_to "https://github.com/HEASARC/cfitsio" "$REPO_PATH" "$REPO_TAG_CFITSIO"

    cd "$REPO_PATH"
    mkdir -p "$LIBCFITSIO_BUILD_FOLDER"
    cd "$LIBCFITSIO_BUILD_FOLDER"
    invoke_build_script_generation "$REPO_PATH"/configure --with-pic --disable-shared --prefix="$INSTALL_DIR"

    make -j "$THREADS"
    make install
}

function install_pybind11 {
    REPO_NAME="pybind11"
    GIT_URL="https://github.com/pybind/pybind11.git"

    install_via_cmake "$REPO_NAME" "$REPO_TAG_PYBIND11" "$GIT_URL" -DBUILD_TESTING=OFF
}


function install_healpix_cxx {
    check_and_install cfitsio

    REPO_NAME="healpix"
    VERSIONED_FOLDER=$REPO_NAME-$REPO_TAG_HEALPIX
    REPO_PATH=$GIT_REPOS/$VERSIONED_FOLDER
    #-------------------------------------------------------------------------------------
    clone_to https://github.com/healpy/healpixmirror.git "$REPO_PATH" "$REPO_TAG_HEALPIX"
    #-------------------------------------------------------------------------------------


    if ! pkg-config --exists "libsharp"; then
        # install libsharp
        LIBSHARP_ROOT=$REPO_PATH/src/common_libraries/libsharp
        LIBSHARP_BUILD_FOLDER=$LIBSHARP_ROOT/build

        cd "$LIBSHARP_ROOT"
        autoreconf -ifv

        mkdir -p "$LIBSHARP_BUILD_FOLDER"
        cd "$LIBSHARP_BUILD_FOLDER"
        invoke_build_script_generation "$LIBSHARP_ROOT"/configure --with-pic --disable-shared --prefix="$INSTALL_DIR"

        make -j "$THREADS"
        make install
    else
        echo "Package libsharp wont be rebuilt as a part of healpix bundle"
    fi

    #-------------------------------------------------------------------------------------
    HEALPIX_ROOT=$REPO_PATH/src/cxx
    HEALPIX_BUILD_DIR=$HEALPIX_ROOT/build

    ld_flags="-L$INSTALL_DIR/lib -Wl,-Bstatic -lcfitsio -Wl,-Bdynamic -lm -lcurl -lz"

    cd "$HEALPIX_ROOT"
    autoreconf -ifv

    mkdir -p "$HEALPIX_BUILD_DIR"
    cd "$HEALPIX_BUILD_DIR"
    INCLUDE_FLAGS="-I$INSTALL_DIR/include"
    invoke_build_script_generation $HEALPIX_ROOT/configure --with-pic --disable-shared --prefix=$INSTALL_DIR CFLAGS="$INCLUDE_FLAGS" CXXFLAGS="$INCLUDE_FLAGS" LDFLAGS="$ld_flags"
    make -j "$THREADS"
    make install

}

function install_yaxt {
    REPO_NAME="yaxt"
    VERSIONED_FOLDER=$REPO_NAME-$REPO_TAG_YAXT
    REPO_PATH=$GIT_REPOS/$VERSIONED_FOLDER
    #-------------------------------------------------------------------------------------
    clone_to https://gitlab.dkrz.de/dkrz-sw/yaxt.git "$REPO_PATH" "$REPO_TAG_YAXT"
    #-------------------------------------------------------------------------------------

    cd "$REPO_PATH"
    autoreconf -ifv

    YAXT_BUILD_FOLDER=$REPO_PATH/build
    mkdir -p "$YAXT_BUILD_FOLDER"
    cd "$YAXT_BUILD_FOLDER"

    invoke_build_script_generation $REPO_PATH/configure --with-pic --disable-mpi-checks --enable-maintainer-mode --with-regard-for-quality=no --without-example-programs --prefix=$INSTALL_DIR

    make -j "$THREADS"
    make install
}

function install_yac {
    install_libfyaml
    install_yaxt

    REPO_NAME="yac"
    VERSIONED_FOLDER=$REPO_NAME-$REPO_TAG_YAC
    REPO_PATH=$GIT_REPOS/$VERSIONED_FOLDER
    #-------------------------------------------------------------------------------------
    clone_to https://gitlab.dkrz.de/dkrz-sw/yac.git "$REPO_PATH" "$REPO_TAG_YAC"
    #-------------------------------------------------------------------------------------

    cd "$REPO_PATH"

    YAC_BUILD_FOLDER=$REPO_PATH/build
    mkdir -p "$YAC_BUILD_FOLDER"
    cd "$YAC_BUILD_FOLDER"

    fyaml_lib_folders="$INSTALL_DIR/lib"
    if [ -d "$INSTALL_DIR/lib64" ]; then
        echo "appending lib64 folder"
        fyaml_lib_folders="$INSTALL_DIR/lib -L$INSTALL_DIR/lib64"
    fi

    invoke_build_script_generation $REPO_PATH/configure --with-pic --enable-openmp --enable-maintainer-mode --disable-mpi-checks --disable-netcdf --prefix=$INSTALL_DIR \
        --with-yaxt-include=$INSTALL_DIR/include \
        --with-yaxt-lib=$INSTALL_DIR/lib \
        --with-fyaml-include=$INSTALL_DIR/include \
        --with-fyaml-lib="$fyaml_lib_folders" \
        FCFLAGS="-fopenmp"

    make -j "$THREADS"
    make install
}


function install_all {
    # echo "========================"
    # echo "== building HEALPIX & Co"
    # check_and_install healpix_cxx
    echo "========================"
    echo "== building YAC & Co"
    check_and_install yac
    echo "========================"
    echo "== building pybind11"
    check_and_install pybind11
    echo "========================"
}


THREADS=8

DEP_TO_BUILD_INSTALL="all" # can be overriden with any particular library to be built

while [[ $# -gt 0 ]]; do
    case "$1" in
        -j)
            THREADS=$2
            shift 2
            ;;
        --dependency)
            DEP_TO_BUILD_INSTALL=$2
            echo "Building the following dependency: $DEP_TO_BUILD_INSTALL"
            shift 2
            ;;
        --prefix_path)
            INSTALL_DIR=$2
            echo "Install prefix path set to: $INSTALL_DIR"
            shift 2
            ;;
        -h|--help)
            usage
            exit 0
            ;;
        *)
            echo "Unknown option: $1"
            usage
            exit 1
            ;;
    esac
done



GIT_REPOS=$(pwd)/external
mkdir -p "$GIT_REPOS"
echo "GIT_REPOS: $GIT_REPOS"

mkdir -p "$INSTALL_DIR"
echo "INSTALL_DIR: $INSTALL_DIR"

install_${DEP_TO_BUILD_INSTALL}
