#include <cassert>
#include <memory>
#include <numeric>

#include "coyoteenv.hpp"
#include <coyote.hpp>

#include <mpi.h>
extern "C" {
#include <yac.h>
}

#include "event.hpp"

namespace coyote {

    std::unique_ptr<CoyoteEnv> CoyoteEnv::_instance = nullptr;

    CoyoteEnv::CoyoteEnv(std::string group_name, uint nthreads)
        : state_def_comps(false), state_sync_def(false), state_enddef(false), _nthreads(nthreads) {
        yac_cdef_calendar(YAC_PROLEPTIC_GREGORIAN);
        if (group_name != "") {
            std::vector<const char*> group_names = {"yac", group_name.data()};
            std::vector<MPI_Comm> comms(2);
            int is_mpi_initialized = -1;
            MPI_Initialized(&is_mpi_initialized);
            if (!is_mpi_initialized) {
                int provided;
                MPI_Init_thread(NULL, NULL, MPI_THREAD_MULTIPLE, &provided);
                assert(provided == MPI_THREAD_MULTIPLE);
                _finalize_mpi = true;
            }
            yac_cmpi_handshake(MPI_COMM_WORLD, 2, group_names.data(), comms.data());
            group_comm = comms[1];
            yac_cinit_comm_instance(comms[0], &yac_instance_id);
        } else {
            yac_cinit_instance(&yac_instance_id);
        }
    }

    CoyoteEnv& CoyoteEnv::getInstance() {
        if (!_instance)
            _instance = std::make_unique<CoyoteEnv>();
        return *_instance;
    }

    int CoyoteEnv::get_yac_instance_id() const { return yac_instance_id; }

    std::string CoyoteEnv::start_datetime() {
        ensure_def_comps();
        return yac_cget_start_datetime_instance(yac_instance_id);
    }

    std::string CoyoteEnv::end_datetime() {
        ensure_def_comps();
        return yac_cget_end_datetime_instance(yac_instance_id);
    }

    MPI_Comm CoyoteEnv::get_group_comm() const { return group_comm; }

    void CoyoteEnv::add_field(std::unique_ptr<Field> field) { _fields.push_back(std::move(field)); }

    void CoyoteEnv::ensure_def_comps() {
        if (state_def_comps)
            return;
        yac_cdef_comps_instance(yac_instance_id, nullptr, 0, nullptr);
        state_def_comps = true;
    }

    void CoyoteEnv::ensure_sync_def() {
        ensure_def_comps();
        if (state_sync_def)
            return;
        yac_csync_def_instance(yac_instance_id);
        state_sync_def = true;
    }

    void CoyoteEnv::ensure_enddef() {
        ensure_sync_def();
        if (state_enddef)
            return;
        yac_cenddef_instance(yac_instance_id);
        state_enddef = true;
    }

    void CoyoteEnv::run() {
        ensure_enddef();

        // run worker threads
        auto do_work = [&]() {
            while (true) {
                try {
                    auto task = _task_queue.dequeue();
                    task();
                } catch (QueueTerminated e) {
                    break;
                }
            };
        };

        assert(_nthreads > 0);
        std::vector<std::thread> worker(_nthreads - 1);
        std::generate(worker.begin(), worker.end(), [&]() { return std::thread(do_work); });

        int done = 0;
        do {
            done = 0;
            std::string curr_datetime =
                (*std::min_element(_fields.begin(), _fields.end(), [](const auto& a, const auto& b) {
                    return a->datetime() < b->datetime();
                }))->datetime();
            for (auto& field : _fields) {
                // skip if done
                if (field->is_done()) {
                    done++;
                    continue;
                }

                // ensure that fields are updated chronologically
                if (field->datetime() != curr_datetime)
                    continue;

                if (!field->is_ready())
                    continue;

                // fire event
                std::shared_ptr<Field> f(field.get(), [](auto p) {
                    p->semaphore.release(); // release the semaphore if all references
                });

                _task_queue.enqueue(
                    [f]() mutable {
                        Event e(f);
                        f->op(Event(e));
                    },
                    {1, f->datetime()});

                // actually we want to offload this also to the
                // task_queue but yac is not thread-safe.
                field->put_forward();
            }
            if (_nthreads == 1 && !_task_queue.empty()) {
                _task_queue.dequeue()();
            }
        } while (done < _fields.size());

        _task_queue.end();

        // help the workers
        do_work();

        for (auto& t : worker)
            t.join();

        // finalization
        _fields.clear();
        yac_cfinalize_instance(yac_instance_id);
        if (_finalize_mpi)
            MPI_Finalize();
    }
} // namespace coyote
